#coding=UTF-8
#synthDrivers/newfon.py
#A part of NonVisual Desktop Access (NVDA)
#Copyright (C) 2006-2018 NVDA Contributors <http://www.nvda-project.org/>
#This file is covered by the GNU General Public License.
#See the file COPYING for more details.

from collections import OrderedDict
from synthDriverHandler import SynthDriver, VoiceInfo, SynthSetting, LanguageInfo, StringParameterInfo,BooleanSynthSetting,NumericSynthSetting

from ctypes import *
import os
import config
import nvwave
import speech
import re
from logHandler import log
file_path = os.path.dirname(os.path.abspath(__file__))

#config
abbreviationsLength = 4

isSpeaking = False
player = None
ProcessAudioCallback = WINFUNCTYPE(c_int, POINTER(c_char),POINTER(c_char),c_int)

@ProcessAudioCallback
def processAudio(udata, buffer,length):
	global isSpeaking,player
	if not isSpeaking: return 1
	player.feed(string_at(buffer, length))
	return 0

re_words = re.compile(r"\b(\w+)\b",re.U)
re_englishLetters = re.compile(r"\b([a-zA-Z])\b")
re_abbreviations = re.compile(ur"\b([bcdfghjklmnpqrstvwxzбвгджзклмнпрстфхцчшщ]{2,})\b",re.U)
re_capAbbreviations = re.compile(ur"([bcdfghjklmnpqrstvwxzбвгджзклмнпрстфхцчшщ]{3,})",re.U|re.I)
re_afterNumber = re.compile(r"(\d+)([^\.\:\-\/\!\?\d])")
re_omittedCharacters = re.compile(r"[\(\)\*_\"]+")
re_zeros = re.compile(r"\b\a?\.?(0+)")

ukrainianRules = {
re.compile(u"\\b(й)\\s",re.U|re.I): U"й",
re.compile(u"\\b(з)\\s",re.U|re.I): U"з",
re.compile(u"\\s(ж)\\b",re.U|re.I): U"ж",
re.compile(u"\\s(б)\\b",re.U|re.I): U"б",
re.compile(ur"'([яюєї])",re.I|re.U): u"ьй\\1",
re.compile(u"ц([ьіяюєї])",re.U|re.I): U"тс\\1"
}

englishLetters = {
'a': u"эй",
'b' : u"би",
'c': u"си",
'd': u"ди",
'e': u"и",
'f': u"эф",
'g': u"джи",
'h': u"эйчь",
'i': u"ай",
'j': u"джей",
'k': u"кей",
'l': u"эл",
'm': u"эм",
'n': u"эн",
'o': u"оу",
'p': u"пи",
'q': u"къю",
'r': u"ар",
's': u"эс",
't': u"ти",
'u': u"ю",
'v': u"ви",
'w': u"да+блъю",
'x': u"экс",
'y': u"вай",
'z': u"зи"
}
russianLetters = {
u"б": u"бэ",
u"в": u"вэ",
u"к": u"ка",
u"с": u"эс",
u"ь": u"мягкий знак",
u"ъ": u"твёрдый знак"
}

englishPronunciation= {
'x': u"кс",
'ee': u"е е",
'e': u"е",
'y': u"ы",
'j': u"дж"
}
#ukrainian to russian character map
#ukrainian soft "g" is not supported, becouse synth does not contain this phonem :(
ukrainianPronunciation = {
u"и": u"ы",
u"і": u"и",
u"ї": u"ййи",
u"е": u"э",
u"є": u"е",
u"ґ": u"г"
}
ukrainianPronunciationOrder = [u"и",u"і", u"ї", u"е", u"є", u"ґ"]

ukrainianLetters = {
u"й": u"йот",
u"ґ": u"Твэрдэ+ гэ",
u"и": u"ы",
u"і": u"и",
u"ї": u"ййи",
u"е": u"э",
u"є": u"е"
}
letters = {}
letters.update(englishLetters)
letters.update(russianLetters)
russianZeros=[u"ноль ",u"ноля",u"нолей"]
ukrainianZeros=[u"нуль ",u"нулі ",u"нулів "]

def subZeros(match,zeros):
	l = len(match.group(1))
	if l == 1: return zeros[0]
	text = " " + str(l) + " "
	l = l%10
	if l == 1: return text+ zeros[0]
	elif l <5: return text+zeros[1]
	else: return text+zeros[2]

def expandAbbreviation(match):
	loweredText = match.group(1).lower()
	l = len(match.group(1))
	if (match.group(1).isupper() and (l <= abbreviationsLength and l > 1) and re_capAbbreviations.match(match.group(1))) or re_abbreviations.match(loweredText):
		expandedText = ""
		for letter in loweredText:
			expandedText += letters[letter] if letters.has_key(letter) else letter
			if letter.isalpha(): expandedText+=" "
		return expandedText
	return loweredText

def subEnglishLetters(match):
	letter = match.group(1).lower()
	return englishLetters[letter]

def preprocessEnglishText(text):
	text = re_englishLetters.sub(subEnglishLetters, text)
	for s in englishPronunciation:
		text = text.replace(s, englishPronunciation[s])
	return text

def preprocessUkrainianText(text):
	for rule in ukrainianRules:
		text = rule.sub(ukrainianRules[rule],text)
	for s in ukrainianPronunciationOrder:
		text = text.replace(s, ukrainianPronunciation[s])
		#stupid python! replace() does not have ignore case, reg exprs also sucks
		text = text.replace(s.upper(), ukrainianPronunciation[s])
	return text

def processText(text,language):
	if len(text) == 1:
		letter = text.lower()
		if language == "uk" and ukrainianLetters.has_key(letter): return ukrainianLetters[letter]
		elif letters.has_key(letter): return letters[letter]
		else: return letter
	text = re_omittedCharacters.sub(" ", text)
	text = re_zeros.sub(lambda match: subZeros(match,russianZeros if language=="ru" else ukrainianZeros),text)
	if language == "uk":
		text = preprocessUkrainianText(text)
	text = re_words.sub(expandAbbreviation,text) #this also lowers the text
	text = preprocessEnglishText(text)
	text = re_afterNumber.sub(r"\1-\2", text)
	return text

sampleRateValues = [10000,11025]
class SynthDriver(SynthDriver):
	name="newfon"
	description = "Newfon"
	supportedSettings=(
		SynthDriver.VoiceSetting(),
		SynthDriver.LanguageSetting(),
		SynthDriver.RateSetting(),
		SynthSetting("accel",_(u"&Ускорение")),
		SynthSetting("sampleRate",_(u"&Частота дискретизации")),
		SynthDriver.PitchSetting(),
		SynthDriver.InflectionSetting(10),
		SynthDriver.VolumeSetting(),
		NumericSynthSetting("pauseBetweenWord",_(u"&Пауза между словами")),
		BooleanSynthSetting("disableSynthDict",_(u"&Выключить встроенный словарь синтезатора (только для русского языка)")),
	)
	_volume = 100
	_language="ru"
	_pitch = 50
	_accel='0'
	_pauseBetweenWord=50
	_disableSynthDict = False
	_sampleRate = '0'
	_inflection=50
	availableVoices = OrderedDict((str(index),VoiceInfo(str(index),name)) for index,name in enumerate([_("male 1"),_("female 1"),_("male 2"),_("female 2")]))
	availableAccels=OrderedDict((str(x), StringParameterInfo(str(x),str(x))) for x in xrange(8))
	global sampleRateValues
	availableSamplerates=OrderedDict((str(s), StringParameterInfo(str(s),str(sampleRateValues[s]))) for s in xrange(2))
	pitchTable=[(90,130),(190,330),(60,120),(220,340)]
	availableLanguages = OrderedDict((("ru",LanguageInfo("ru")), ("uk",LanguageInfo("uk"))))
	newfon_lib = None
	sdrvxpdbDll = None
	dictDll = None

	@classmethod
	def check(cls):
		return os.path.isfile(os.path.join(file_path, 'newfon_nvda.dll'))

	def calculateMinMaxPitch(self,pitch,inflection):
		min,max=self.pitchTable[int(self.voice)]
		i=max-min
		i=int((i/50.0)*((inflection-50)/2))
		min-=i
		max+=i
		i=int((pitch-50)/1.3)
		min+=i
		max+=i
		return min,max

	def __init__(self):
		global player
		player = nvwave.WavePlayer(channels=1, samplesPerSec=10000, bitsPerSample=8, outputDevice=config.conf["speech"]["outputDevice"])
		self.hasDictLib = os.path.isfile(os.path.join(file_path, 'dict.dll'))
		if self.hasDictLib:
			self.sdrvxpdb_lib = windll.LoadLibrary(os.path.join(file_path, 'sdrvxpdb.dll'))
			self.dict_lib = windll.LoadLibrary(os.path.join(file_path, 'dict.dll'))
		self.newfon_lib = windll.LoadLibrary(os.path.join(file_path, 'newfon_nvda.dll'))
		self.newfon_lib.speakText.argtypes = [c_char_p, c_int]
		if not self.newfon_lib.initialize(): raise Exception
		self.newfon_lib.set_callback(processAudio)
		self.newfon_lib.set_dictionary(1)
		self.rate = 70

	def terminate(self):
		self.cancel()
		global player
		player.close()
		player=None
		self.newfon_lib.terminate()
		del self.newfon_lib
		if self.hasDictLib:
			del self.dict_lib
			del self.sdrvxpdb_lib
	def speak(self,speechSequence):
		global isSpeaking
		lastIndex = None
		isSpeaking = True
		textList=[]
		for item in speechSequence:
			if isinstance(item,basestring):
				textList.append(processText(item,self._language))
			elif isinstance(item,speech.IndexCommand):
				lastIndex=item.index
			elif isinstance(item,speech.SpeechCommand):
				log.debugWarning("Unsupported speech command: %s"%item)
			else:
				log.error("Unknown speech: %s"%item)

		text="".join(textList)
		text = text.encode('cp1251', errors='replace')
		if lastIndex is not  None:
			self.newfon_lib.speakText(text,lastIndex)
			self.newfon_lib.set_pause(int(self._pauseBetweenWord))
		else:
			self.newfon_lib.speakText(text,-1)
			self.newfon_lib.set_pause(int(self._pauseBetweenWord))



#	def speakText(self, text, index=None):
## This method is deprecated
#		global isSpeaking
#		isSpeaking = True
#		text = processText(text, self._language)
#		text = text.encode('cp1251', errors='replace')
#		if index is not None: 
#			self.newfon_lib.speakText(text,index)
#			self.newfon_lib.set_pause(int(self._pauseBetweenWord))
#		else:
#			self.newfon_lib.speakText(text,-1)
#			self.newfon_lib.set_pause(int(self._pauseBetweenWord))

	def _get_lastIndex(self):
		return self.newfon_lib.get_lastIndex()

	def cancel(self):
		self.newfon_lib.cancel()
		global isSpeaking,player
		isSpeaking = False
		player.stop()

	def _get_voice(self):
		return str(self.newfon_lib.get_voice())

	def _set_voice(self, value):
		self.newfon_lib.set_voice(int(value))
		self._set_pitch(self._pitch)

	def _get_volume(self):
		return self._volume

	def _set_volume(self,value):
		self.newfon_lib.set_volume(value)
		self._volume = value

	def _get_rate(self):
		return self._rate

	def _set_rate(self, value):
		self.newfon_lib.set_rate(value)
		self._rate = value

	def _set_pitch(self, value):
		#if value <= 50: value = 50
		#self.newfon_lib.set_accel(value/5 -10 )
		self._pitch = value
		min,max=self.calculateMinMaxPitch(self._pitch,self._inflection)
		self.newfon_lib.set_pitch_min(min)
		self.newfon_lib.set_pitch_max(max)

	def _get_pitch(self):
		return self._pitch

	def pause(self, switch):
		global player
		player.pause(switch)

	def _get_language(self):
		return self._language

	def _set_language(self, language):
		self._language = language
		self._set_disableSynthDict(self._disableSynthDict) 

	def _set_inflection(self,inflection):
		self._inflection=inflection
		self._set_pitch(self._pitch)

	def _get_inflection(self):
		return self._inflection

	def _set_accel(self,a):
		self._accel=a
		self.newfon_lib.set_accel(int(a))

	def _get_accel(self):
		return self._accel

	def _set_sampleRate(self,a):
		global player
		self._sampleRate=a
		self.cancel()
		player.close()
		player = nvwave.WavePlayer(channels=1, samplesPerSec=sampleRateValues[int(a)], bitsPerSample=8, outputDevice=config.conf["speech"]["outputDevice"])

	def _get_sampleRate(self):
		return self._sampleRate

	def _set_pauseBetweenWord(self,a):
		self._pauseBetweenWord=a

	def _get_pauseBetweenWord(self):
		return self._pauseBetweenWord

	def _set_disableSynthDict(self,a):
		if not self.hasDictLib: return
		self._disableSynthDict=a
		if self.language == "ru":
			if self._disableSynthDict == False:
				self.newfon_lib.set_dictionary(1)
			elif self._disableSynthDict == True:
				self.newfon_lib.set_dictionary(0)
		else:
			self.newfon_lib.set_dictionary(0)

	def _get_disableSynthDict(self):
		return self._disableSynthDict
